/*
* Titel........ nCH wireless sensor based on Hope RF's RFM02 transmitter
* Version...... 00
* Dateiname.... main.c
* erstellt am.. 2016-10-30
* gendert am.. 2017-05-02
* Platform..... ATMEL ATtiny261A, external 2MHz clock
* Autor........ Clemens B Fahr
* email........ clemens.fahr[at]fahrnet.de
* Kommentar.... based on RFM_wireless_4CH_sensor_02_mod
* .............	modify init delay after POR of radio module from 300 to 200 ms
* .............	speed up settling time for VADC to settle (from 100ms to 40ms)
* ............. change measurement cycle time from 30sec to 90sec
* ............. make number of measurement channels configurable (compile time)
* ............. automatic selection of MQTT topic ID; minor touch up and code review
* ............. !!! set compiler optimisation level to "s" !!!
* ............. modify the MAIN_tresholdExceed function, simplify parameter interface
* ............. wakeup issue bug-fixed: BUGFIX: need to clear/ set the wakeup timer flag 
* ............. in the power command after each wake up from the RFM !!!
* ............. implement sensor option 3channel water level detection; add a threshold based
* ............. TX filter to limit number of telegrams being sent if there is no change
*/

// includes
//------------------------------------------------------------------
#include "global.h"
#include "lowlevel.h"
#include "main.h"

// interrupt service routines -- ISRs
//------------------------------------------------------------------
// mind spelling SIGNAL command in capitals only
// all interrupt sources, though not used, should be captured


ISR(TIMER1_OVF_vect)
{
        // only catches the interrupt - used in the LOWLEVEL_delay_sleep_100ms routine
		;
}

ISR(INT0_vect)
{
	// disable the INT0 interrupt source and clear flag
	GIMSK &= ~(0x01 << INT0);
	GIFR = (0x01 << INTF0);
	
	// clear the last nIRQ from radio module by a status register read command
	//LOWLEVEL_status_register_read();	// seems to be causing the radio SoC not responding any more 2017-02-21
        
    if ( REPEAT_ACTIVE ) {
		if ( (status_flag & 0x00FF) == FIRST_RF_TX) {
			status_flag &= 0xFF00;
			status_flag |= REPEAT_RF_TX;
		}
		else if ( (status_flag & 0x00FF) == REPEAT_RF_TX) {
			status_flag &= 0xFF00;
			status_flag |= FIRST_RF_TX;
		}
	}
	
	if ( (status_flag & 0xFF00) < REPEAT_INIT_ALL ) {
		status_flag += 0x0100; // increment the upper byte
	}
	else {
		status_flag &= 0x00FF; 	// clear the init counter
	}
}

//------------------------------------------------------------------

uint8_t MAIN_tresholdExceed(uint16_t current_value, uint8_t elem, uint8_t threshold_value, uint8_t TX_holdoff)
// compares the current_value to a previous value from ADC_previous[elem..elem+1] and returns TRUE if the absolute difference is
// greater than given through threshold_value
// for TX_holdoff == 0, the holdoff for the respective ADC-element is disabled
{
        int16_t delta_value;
        uint8_t return_value;
		
		return_value = TRUE;
        
        delta_value = abs(current_value - ADC_previous_value[elem]);
		
		#ifdef DEBUGADCTHRESHOLD
		LOWLEVEL_debugOut(0xB0 | elem);
		LOWLEVEL_debugOut(elem);
		LOWLEVEL_debugOut((uint8_t)delta_value);
		#endif
        
        if ( delta_value >= threshold_value ) {
                return_value = TRUE;
                ADC_previous_value[elem] = current_value;
                if (TX_holdoff > 0) {
                        ADC_TX_holdoff[elem] = ADC_TX_holdoff[elem] | TRUE;
                }
        }
        else {
                if (TX_holdoff > 0) {
                        ADC_TX_holdoff[elem] = ADC_TX_holdoff[elem] | FALSE;
                }
                return_value = FALSE;
        }
        
        if (TX_holdoff > 0) {
                if ( global_TX_holdoff < TX_holdoff ) {
                        return_value = FALSE;
                }
                else if ( (global_TX_holdoff == TX_holdoff) &  ADC_TX_holdoff[elem] ) {
                        ADC_TX_holdoff[elem] = FALSE;
                        return_value = TRUE;
                }
        }
		
		// for debug purposes
		#ifdef ADCTHRESHOLDINACTIVE
		return_value = TRUE;
		#endif
		
		#ifdef DEBUGADCTHRESHOLD
		LOWLEVEL_debugOut(0xB0 | elem);
		LOWLEVEL_debugOut(elem);
		LOWLEVEL_debugOut(return_value);
		#endif
        
        return return_value;
}

uint8_t MAIN_buildMessage(void)
{
	/* create the telegram as MQTT-S type PUBLISH telegram, hence MsgType will be 0x0C
	*  with QoS -1, the MsgID will be always zero
	*
	* MQTTS |          |        |   |      |     | Len | MsgType | Flg | TpcID | MsgID  | Data |     |           |
	* byte  |   0-2    |  3-4   | 5 |  6-7 | 8-9 | 10  |    11   |  12 | 13-14 | 15-16  | 17:n | n+1 |  n+2:n+4  |
	* reps  | PREAMBLE |  RFM   | L |  ID  | UID |     |         |     |       |        |      | CRC | POSTAMBLE |
	* ------+----------+--------+---+------+-----+-----+---------+-----+-------+--------+------+-----------------+
	* value | 0xAAAAAA | 0x2DD4 |   |      |     |     |   0x0C  |     |       | 0x0000 |      |     | 0xAAAAAA  |
	*
	* - L is 1byte payload length starting byte 5 up to incl. byte n+1
	* - ID is 2byte common identifier for the TX-node
	* - UID is 2byte unique identifiers for the TX-node
	* - CRC is the modular sum of bytes 5 to incl. n
	*/
	
	/* As we are using a pre-defined topic ID, the TopicIdType in the MQTT-S Flg (Flag byte) will be 0b01
	*  topic ID (tpcID) is selected automatically according to definition:
	*	temperature | temp. external 1 | external 2 | illumination 
	*	(topic ID value is or-ed from sensor-setting to form the first byte, second byte represents the 
	*	bit-reversed pattern from first byte)
	*  Len is the 1byte MQTT-S length info including the Len byte itself.
	*  the Data bytes are defined as follows:
	*
	* byte  | 17-18 | +2byte | +2byte | +2byte | +2byte | NOTE1
	* reps  |  BATT | [TEMP] | [ILLU] | [EXT1] | [EXT2] |
	* ------+-------+--------+--------+--------+--------+
	*
	* NOTE1 depends on setting in global.h; BATT is not optional
	*
	* alternatively (as 3CH water level sensor)
	*
	* byte  | 17-18 | +1byte | +1byte |
	* reps  |  BATT | WTRLVL | ERRFLG |
	* ------+-------+--------+--------+
	*
	*
	*/
	
	uint16_t ADC_value;
	uint8_t i, j, length_byte=0, length_byte2=0, start_CRC, stop_CRC, topic, topic_ID=0, aux_value, error_flag=OFF;
	uint16_t CRC_byte=0;
	
	// preamble
	i=0;
	TX_buffer[i] = 0xAA;
	TX_buffer[++i] = 0xAA;
	TX_buffer[++i] = 0xAA;
	//TX_buffer[++i] = 0xAA;
	//TX_buffer[++i] = 0xAA;
	TX_buffer[++i] = 0x2D;
	TX_buffer[++i] = 0xD4;
	
	// telegram length L - will be added at the end; this creates the spacer and inits with default value
	TX_buffer[++i] = 0xFF;
	length_byte = i; // should be position 6, array index 5
	start_CRC = i;
	
	// add device ID - 2byte
	TX_buffer[++i] = eeprom_read_byte(&DEVICE_ID2); // common device ID2
	TX_buffer[++i] = eeprom_read_byte(&DEVICE_ID1); // common device ID1
	
	// add device UID - 2byte
	TX_buffer[++i] = eeprom_read_byte(&DEVICE_UID2); // unique device ID2
	TX_buffer[++i] = eeprom_read_byte(&DEVICE_UID1); // unique device ID1
	
	#ifdef DEBUGADCON
	LOWLEVEL_debugOut(0xB1);
	LOWLEVEL_debugOut(i);
	#endif
	
	// telegram length Len as of MQTT-S - will be added at the end; this creates the spacer and inits with default value
	TX_buffer[++i] = 0xFF;
	length_byte2 = i; // should be position 11, array index 10
	
	// message type as of MQTT-S - PUBLISH
	TX_buffer[++i] = 0x0C;
	
	// flag as of MQTT-S - QoS is -1 => 0b11 = 3; Retain is active => 0b1 = 1; TopicIdType is pre-defined => 0b01 = 1
	TX_buffer[++i] = (3 << 5) | (1 << 4) | 1;
	
	// Topic ID is generated upon measurement channel config - added at the end, inits with defaults
	TX_buffer[++i] = 0x00;
	topic = i;
	TX_buffer[++i] = 0x00;
	
	// Message ID is 0x0000 - due to the nature of this QoS
	TX_buffer[++i] = 0x00;
	TX_buffer[++i] = 0x00;
	
	// add sensor and battery value hereafter each uint16 values
	// power up the sensors and AVREF
	// toggle respective port pin
	SENS_EN(LOW);
	// wait for the voltage at AREF to settle
	LOWLEVEL_delay_1ms(40); 
	
	#ifdef DEBUGADCON
	LOWLEVEL_debugOut(0xB2);
	LOWLEVEL_debugOut(i);
	#endif
	
    ADC_change = FALSE;
	
	j=0;
	ADC_value = LOWLEVEL_adcConvert(BATT_SENS_MUX,0);
	TX_buffer[++i] = (uint8_t) ADC_value; 		// battery LSB
	TX_buffer[++i] = ADC_value >> 8; 	// battery MSB
    ADC_change = MAIN_tresholdExceed(ADC_value, j, 10, MAX_TX_HOLDOFF) | ADC_change; // 10 digits equivalent to 50mV
	
	#ifdef DEBUGADCON
	LOWLEVEL_debugOut(0xB3);
	LOWLEVEL_debugOut(i);
	#endif
	
	#ifdef S_TEMP
	
		j += 1;
		ADC_value = LOWLEVEL_adcConvert(TEMP_SENS_MUX,1);
		TX_buffer[++i] = (uint8_t) ADC_value; 		// NTC temp sensor LSB
		TX_buffer[++i] = ADC_value >> 8; 	// NTC temp sensor MSB
		ADC_change = MAIN_tresholdExceed(ADC_value, j, 3, 0) | ADC_change; // 3 digits equivalent to 0.75K to 0.375K
		topic_ID = topic_ID | S_TEMP;
	
		#ifdef DEBUGADCON
		LOWLEVEL_debugOut(0xB4);
		LOWLEVEL_debugOut(i);
		#endif
	
	#endif
	
	#ifdef S_WATER3CH
	
		error_flag = OFF;
		// channel 1
		ADC_value = LOWLEVEL_adcConvert(WATERLEVELCH1_SENS_MUX,1);
		aux_value = 0x00;
		if (ADC_value <= 755) {
			aux_value = 1;
		}
		// channel 2
		ADC_value = LOWLEVEL_adcConvert(WATERLEVELCH2_SENS_MUX,1);
		if ( (ADC_value <= 755) & (aux_value == 1) ) {
			aux_value = 2;
		}
		if ( (ADC_value <= 755) & (aux_value == 0) ) {
			error_flag = ON;
		}
		// channel 3
		ADC_value = LOWLEVEL_adcConvert(WATERLEVELCH3_SENS_MUX,1);
		if ( (ADC_value <= 755) & (aux_value == 2) ) {
			aux_value = 3;
		}
		if ( (ADC_value <= 755) & (aux_value <= 1) ) {
			error_flag = ON;
		}
		TX_buffer[++i] = aux_value;		// water level
		TX_buffer[++i] = error_flag; 	// sensor error flag
		j += 1;
		ADC_change = MAIN_tresholdExceed(aux_value, j, 1, 3) | ADC_change; // returns true if level changes by one
		j += 1;
		ADC_change = MAIN_tresholdExceed(error_flag, j, 1, 3) | ADC_change; // returns true if error_flag toggles
		topic_ID = topic_ID | S_WATER3CH;
		
	#endif
	
	#ifdef S_ILLU
   
		j += 1;
		ADC_value = LOWLEVEL_adcConvert(ILLUM_SENS_MUX,1);
		TX_buffer[++i] = (uint8_t) ADC_value; 		// illumination sensor LSB
		TX_buffer[++i] = ADC_value >> 8; 	// illumination sensor MSB
		ADC_change = MAIN_tresholdExceed(ADC_value, j, 28, 5) | ADC_change; // 28 digits empirically determined, illumination curve is logarithmic
		topic_ID = topic_ID | S_ILLU;
	
		#ifdef DEBUGADCON
		LOWLEVEL_debugOut(0xB5);
		LOWLEVEL_debugOut(i);
		#endif
	
	#endif
	
	#ifdef S_TEMP_EXT1
    
		j += 1;
		ADC_value = LOWLEVEL_adcConvert(EXT_SENS1_MUX,1);
		TX_buffer[++i] = (uint8_t) ADC_value; 		// external NTC temp sensor1 LSB
		TX_buffer[++i] = ADC_value >> 8; 	// external NTC temp sensor1 MSB
		ADC_change = MAIN_tresholdExceed(ADC_value, j, 4, 5) | ADC_change; // 3 digits equivalent to 0.75K to 0.375K
		topic_ID = topic_ID | S_TEMP_EXT1;
	
		#ifdef DEBUGADCON
		LOWLEVEL_debugOut(0xB6);
		LOWLEVEL_debugOut(i);
		#endif
	
	#endif
	
	#ifdef S_TEMP_EXT2
	
		j += 1;
		ADC_value = LOWLEVEL_adcConvert(EXT_SENS2_MUX,1);
		TX_buffer[++i] = (uint8_t) ADC_value; 		// external NTC temp sensor2 LSB
		TX_buffer[++i] = ADC_value >> 8; 	// external NTC temp sensor2 MSB
		ADC_change = MAIN_tresholdExceed(ADC_value, j, 4, 5) | ADC_change; // 3 digits equivalent to 0.75K to 0.375K
		topic_ID = topic_ID | S_TEMP_EXT2;
	
		#ifdef DEBUGADCON
		LOWLEVEL_debugOut(0xB7);
		LOWLEVEL_debugOut(i);
		#endif
	
	#endif
	
	stop_CRC = i;
	
	// power down the sensors and AVREF
	// toggle respective port pin
	SENS_EN(HIGH);
	
	// correct the MQTT topic ID
	TX_buffer[topic+1] = topic_ID;
	// create revert image of ID
	for (j=0;j<8;j++) {
		if (topic_ID & 0x01) {
			TX_buffer[topic] |= (0x01 << (7-j));	
		}
		topic_ID = topic_ID >> 1;
	}
	
	// correct the MQTT-S Len
	TX_buffer[length_byte2] = i-length_byte2+1;
	
	// correct the complete telegram's length L
	TX_buffer[length_byte] = i-length_byte+2;
	
	// calculate the CRC_byte
	CRC_byte = 0;
	for (uint8_t j=start_CRC;j<=stop_CRC;j++) {
		CRC_byte += TX_buffer[j];
	}
	// the modular sum checksum of all the ADC values incl. the length byte
	TX_buffer[++i] = (uint8_t) (CRC_byte & 0x00FF);
	
	// post-amble
	TX_buffer[++i] = 0xAA;
	TX_buffer[++i] = 0xAA;
	TX_buffer[++i] = 0xAA;
	
	i++;
	
	#ifdef DEBUGADCON
	LOWLEVEL_debugOut(0xB8);
	LOWLEVEL_debugOut(i);
	#endif
	
	return i;
}

//------------------------------------------------------------------

int main(void)
{

	uint8_t j;
	uint16_t CRC_byte;

	status_flag = FIRST_RF_TX;

    ADC_change = TRUE;
    global_TX_holdoff = 0;
	
	// run all low level initialisations for the very first time
	LOWLEVEL_init_all();  
	// initialise the RFM module
	LOWLEVEL_init_RFM();
	
	// init ADC buffer; used determine the deviation from current to previous measurement
    LOWLEVEL_init_ADC_buffer(MAX_ADC_BUFFER); 
	
	#ifdef DEBUGON
	LOWLEVEL_debugOut(0x01);
	#endif
	
	LOWLEVEL_delay_1ms(300); // need to wait for the RFM module to be ready after POR
	
	#ifdef DEBUGON
	LOWLEVEL_debugOut(0x02);
	#endif
	
	while  (1) {
	
		if ( (status_flag & 0xFF00) == REPEAT_INIT_ALL) {
			#ifdef DEBUGON
			LOWLEVEL_debugOut(0x03);
			#endif
			// run all low level initialisations
			LOWLEVEL_init_all();
			// initialise the RFM module
			LOWLEVEL_init_RFM();
			#ifdef DEBUGON
			LOWLEVEL_debugOut(0xA0);
			LOWLEVEL_debugOut((uint8_t) status_flag);
			LOWLEVEL_debugOut(status_flag >> 8);
			#endif
		}
		
		if ( (status_flag & 0x00FF) == FIRST_RF_TX) {
			// toggle the LED to indicate activity
			#ifndef DEBUGON
			LOWLEVEL_blink_status_LED(1);
			#endif
			#ifdef DEBUGON
			LOWLEVEL_debugOut(0xB0);
			#endif
			// prepare the telegram by measuring and creating CRC
			telegram_length = MAIN_buildMessage();
			#ifdef DEBUGON
			LOWLEVEL_debugOut(0xD0);
			LOWLEVEL_debugOut(telegram_length);
			#endif
			//only send telegram, if any value change has occurred since last measurement
            if ( ADC_change | (global_TX_holdoff >= MAX_TX_HOLDOFF) ) {
				// send the telegram with the number of bytes prepared
				LOWLEVEL_send_telegram_RFM(telegram_length);
                // toggle the LED to indicate activity
				#ifndef DEBUGON
				LOWLEVEL_blink_status_LED(2);
				#endif
                global_TX_holdoff = 0;
				#ifdef DEBUGON
				LOWLEVEL_debugOut(0xE0);
				//LOWLEVEL_debugOut(global_TX_holdoff);
				#endif
            }
            else {
                global_TX_holdoff += 1;
				#ifdef DEBUGON
				LOWLEVEL_debugOut(0xF0);
				#endif
            }
			// go to sleep and wait for the interrupt of the radio module to proceed (this is the sleep repeat)
			if ( REPEAT_ACTIVE ) {
				LOWLEVEL_sleep_RFM(REPEAT_SLEEP);
				#ifdef DEBUGON
				LOWLEVEL_debugOut(0xF1);
				#endif
			}
            else {
				LOWLEVEL_sleep_RFM(LONG_SLEEP);
				#ifdef DEBUGON
				LOWLEVEL_debugOut(0xF2);
				#endif
            }
		}
		else if ( (status_flag & 0x00FF) == REPEAT_RF_TX) {
			
			// NEED MDOFICATION HERE IF TO BE USED!!!!
			
			// set the DUP flag in the TX-buffer
			TX_buffer[12] |= (1 << 7);
			// re-calculate the CRC
			CRC_byte = 0;
			for (j=5;j<=26;j++) {
				CRC_byte = CRC_byte + TX_buffer[j];
			}
			TX_buffer[27] = CRC_byte;
			// initialise the RFM module
			LOWLEVEL_init_RFM();
			// send the telegram with the number of bytes prepared
			LOWLEVEL_send_telegram_RFM(telegram_length);
            // toggle the LED to indicate activity
            LOWLEVEL_blink_status_LED(2);
            global_TX_holdoff = 0;
			// go to sleep and wait for the interrupt of the radio module to proceed (this is the long sleep)
			LOWLEVEL_sleep_RFM(LONG_SLEEP);
		}
		
		// enable to wake up on INT0 event, need low level interrupt for INT0 
		// only level interrupt is supported when power down is issued as sleep mode
		GIMSK |= (0x01 << INT0);
		sei();
		#ifdef DEBUGON
		LOWLEVEL_debugOut(0xF3);
		LOWLEVEL_debugOut((uint8_t) status_flag);
		LOWLEVEL_debugOut(status_flag >> 8);
		#endif
		// enable the sleep mode function, set the sleep mode to POWER DOWN
		set_sleep_mode(SLEEP_MODE_PWR_DOWN);
		sleep_mode();
		// clear the last nIRQ from radio module by a status register read command
		CRC_byte = LOWLEVEL_status_register_read(); // missusing the CRC-byte here for debug purposes
		// entry point after repeat-sleep wakeup through external interrupt INT0 is here
		cli();
		#ifdef DEBUGON
		LOWLEVEL_debugOut(0x10);
		LOWLEVEL_debugOut((uint8_t) status_flag);
		LOWLEVEL_debugOut(status_flag >> 8);
		LOWLEVEL_debugOut((uint8_t) CRC_byte);
		LOWLEVEL_debugOut(CRC_byte >> 8);
		LOWLEVEL_debugOut(global_TX_holdoff);
		#endif
	}
	
	return 0;  //never reached
}
